"""
To easily verify if our normalizer raises the right error codes, just use the
tests of pydocstyle.
"""

import difflib
import re
from functools import total_ordering
from typing import Iterator, Tuple

import parso
from parso.utils import python_bytes_to_unicode


@total_ordering
class WantedIssue:
    def __init__(self, code: str, line: int, column: int) -> None:
        self.code = code
        self._line = line
        self._column = column

    def __eq__(self, other):
        return self.code == other.code and self.start_pos == other.start_pos

    def __lt__(self, other: 'WantedIssue') -> bool:
        return self.start_pos < other.start_pos or self.code < other.code

    def __hash__(self) -> int:
        return hash(str(self.code) + str(self._line) + str(self._column))

    @property
    def start_pos(self) -> Tuple[int, int]:
        return self._line, self._column


def collect_errors(code: str) -> Iterator[WantedIssue]:
    for line_nr, line in enumerate(code.splitlines(), 1):
        match = re.match(r'(\s*)#: (.*)$', line)
        if match is not None:
            codes = match.group(2)
            for code in codes.split():
                code, _, add_indent = code.partition(':')
                column = int(add_indent or len(match.group(1)))

                code, _, add_line = code.partition('+')
                ln = line_nr + 1 + int(add_line or 0)

                yield WantedIssue(code[1:], ln, column)


def test_normalizer_issue(normalizer_issue_case):
    def sort(issues):
        issues = sorted(issues, key=lambda i: (i.start_pos, i.code))
        return ["(%s, %s): %s" % (i.start_pos[0], i.start_pos[1], i.code)
                for i in issues]

    with open(normalizer_issue_case.path, 'rb') as f:
        code = python_bytes_to_unicode(f.read())

    desired = sort(collect_errors(code))

    grammar = parso.load_grammar(version=normalizer_issue_case.python_version)
    module = grammar.parse(code)
    issues = grammar._get_normalizer_issues(module)
    actual = sort(issues)

    diff = '\n'.join(difflib.ndiff(desired, actual))
    # To make the pytest -v diff a bit prettier, stop pytest to rewrite assert
    # statements by executing the comparison earlier.
    _bool = desired == actual
    assert _bool, '\n' + diff



"""
Test all things related to the ``jedi.cache`` module.
"""

import os
import pytest
import time
from pathlib import Path

from parso.cache import (_CACHED_FILE_MAXIMUM_SURVIVAL, _VERSION_TAG,
                         _get_cache_clear_lock_path, _get_hashed_path,
                         _load_from_file_system, _NodeCacheItem,
                         _remove_cache_and_update_lock, _save_to_file_system,
                         load_module, parser_cache, try_to_save_module)
from parso._compatibility import is_pypy
from parso import load_grammar
from parso import cache
from parso import file_io
from parso import parse

skip_pypy = pytest.mark.skipif(
    is_pypy,
    reason="pickling in pypy is slow, since we don't pickle,"
           "we never go into path of auto-collecting garbage"
)


@pytest.fixture()
def isolated_parso_cache(monkeypatch, tmpdir):
    """Set `parso.cache._default_cache_path` to a temporary directory
    during the test. """
    cache_path = Path(str(tmpdir), "__parso_cache")
    monkeypatch.setattr(cache, '_default_cache_path', cache_path)
    return cache_path


def test_modulepickling_change_cache_dir(tmpdir):
    """
    ParserPickling should not save old cache when cache_directory is changed.

    See: `#168 <https://github.com/davidhalter/jedi/pull/168>`_
    """
    dir_1 = Path(str(tmpdir.mkdir('first')))
    dir_2 = Path(str(tmpdir.mkdir('second')))

    item_1 = _NodeCacheItem('bla', [])
    item_2 = _NodeCacheItem('bla', [])
    path_1 = Path('fake path 1')
    path_2 = Path('fake path 2')

    hashed_grammar = load_grammar()._hashed
    _save_to_file_system(hashed_grammar, path_1, item_1, cache_path=dir_1)
    parser_cache.clear()
    cached = load_stored_item(hashed_grammar, path_1, item_1, cache_path=dir_1)
    assert cached == item_1.node

    _save_to_file_system(hashed_grammar, path_2, item_2, cache_path=dir_2)
    cached = load_stored_item(hashed_grammar, path_1, item_1, cache_path=dir_2)
    assert cached is None


def load_stored_item(hashed_grammar, path, item, cache_path):
    """Load `item` stored at `path` in `cache`."""
    item = _load_from_file_system(hashed_grammar, path, item.change_time - 1, cache_path)
    return item


@pytest.mark.usefixtures("isolated_parso_cache")
def test_modulepickling_simulate_deleted_cache(tmpdir):
    """
    Tests loading from a cache file after it is deleted.
    According to macOS `dev docs`__,

        Note that the system may delete the Caches/ directory to free up disk
        space, so your app must be able to re-create or download these files as
        needed.

    It is possible that other supported platforms treat cache files the same
    way.

    __ https://developer.apple.com/library/content/documentation/FileManagement/Conceptual/FileSystemProgrammingGuide/FileSystemOverview/FileSystemOverview.html
    """  # noqa
    grammar = load_grammar()
    module = 'fake parser'

    # Create the file
    path = Path(str(tmpdir.dirname), 'some_path')
    with open(path, 'w'):
        pass
    io = file_io.FileIO(path)

    try_to_save_module(grammar._hashed, io, module, lines=[])
    assert load_module(grammar._hashed, io) == module

    os.unlink(_get_hashed_path(grammar._hashed, path))
    parser_cache.clear()

    cached2 = load_module(grammar._hashed, io)
    assert cached2 is None


def test_cache_limit():
    def cache_size():
        return sum(len(v) for v in parser_cache.values())

    try:
        parser_cache.clear()
        future_node_cache_item = _NodeCacheItem('bla', [], change_time=time.time() + 10e6)
        old_node_cache_item = _NodeCacheItem('bla', [], change_time=time.time() - 10e4)
        parser_cache['some_hash_old'] = {
            '/path/%s' % i: old_node_cache_item for i in range(300)
        }
        parser_cache['some_hash_new'] = {
            '/path/%s' % i: future_node_cache_item for i in range(300)
        }
        assert cache_size() == 600
        parse('somecode', cache=True, path='/path/somepath')
        assert cache_size() == 301
    finally:
        parser_cache.clear()


class _FixedTimeFileIO(file_io.KnownContentFileIO):
    def __init__(self, path, content, last_modified):
        super().__init__(path, content)
        self._last_modified = last_modified

    def get_last_modified(self):
        return self._last_modified


@pytest.mark.parametrize('diff_cache', [False, True])
@pytest.mark.parametrize('use_file_io', [False, True])
def test_cache_last_used_update(diff_cache, use_file_io):
    p = Path('/path/last-used')
    parser_cache.clear()  # Clear, because then it's easier to find stuff.
    parse('somecode', cache=True, path=p)
    node_cache_item = next(iter(parser_cache.values()))[p]
    now = time.time()
    assert node_cache_item.last_used <= now

    if use_file_io:
        f = _FixedTimeFileIO(p, 'code', node_cache_item.last_used - 10)
        parse(file_io=f, cache=True, diff_cache=diff_cache)
    else:
        parse('somecode2', cache=True, path=p, diff_cache=diff_cache)

    node_cache_item = next(iter(parser_cache.values()))[p]
    assert now <= node_cache_item.last_used <= time.time()


@skip_pypy
def test_inactive_cache(tmpdir, isolated_parso_cache):
    parser_cache.clear()
    test_subjects = "abcdef"
    for path in test_subjects:
        parse('somecode', cache=True, path=os.path.join(str(tmpdir), path))
    raw_cache_path = isolated_parso_cache.joinpath(_VERSION_TAG)
    assert raw_cache_path.exists()
    dir_names = os.listdir(raw_cache_path)
    a_while_ago = time.time() - _CACHED_FILE_MAXIMUM_SURVIVAL
    old_paths = set()
    for dir_name in dir_names[:len(test_subjects) // 2]:  # make certain number of paths old
        os.utime(raw_cache_path.joinpath(dir_name), (a_while_ago, a_while_ago))
        old_paths.add(dir_name)
    # nothing should be cleared while the lock is on
    assert _get_cache_clear_lock_path().exists()
    _remove_cache_and_update_lock()  # it shouldn't clear anything
    assert len(os.listdir(raw_cache_path)) == len(test_subjects)
    assert old_paths.issubset(os.listdir(raw_cache_path))

    os.utime(_get_cache_clear_lock_path(), (a_while_ago, a_while_ago))
    _remove_cache_and_update_lock()
    assert len(os.listdir(raw_cache_path)) == len(test_subjects) // 2
    assert not old_paths.intersection(os.listdir(raw_cache_path))


@skip_pypy
def test_permission_error(monkeypatch):
    def save(*args, **kwargs):
        nonlocal was_called
        was_called = True
        raise PermissionError

    was_called = False

    monkeypatch.setattr(cache, '_save_to_file_system', save)
    try:
        with pytest.warns(Warning):
            parse(path=__file__, cache=True, diff_cache=True)
        assert was_called
    finally:
        parser_cache.clear()



from __future__ import absolute_import

from enum import Enum


class TokenType:
    name: str
    contains_syntax: bool

    def __init__(self, name: str, contains_syntax: bool = False):
        self.name = name
        self.contains_syntax = contains_syntax

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.name)


class PythonTokenTypes(Enum):
    STRING = TokenType('STRING')
    NUMBER = TokenType('NUMBER')
    NAME = TokenType('NAME', contains_syntax=True)
    ERRORTOKEN = TokenType('ERRORTOKEN')
    NEWLINE = TokenType('NEWLINE')
    INDENT = TokenType('INDENT')
    DEDENT = TokenType('DEDENT')
    ERROR_DEDENT = TokenType('ERROR_DEDENT')
    FSTRING_STRING = TokenType('FSTRING_STRING')
    FSTRING_START = TokenType('FSTRING_START')
    FSTRING_END = TokenType('FSTRING_END')
    OP = TokenType('OP', contains_syntax=True)
    ENDMARKER = TokenType('ENDMARKER')



#: E251:11 E251:13
def foo(bar = False):
    '''Test function with an error in declaration'''
    pass


#: E251:8
foo(bar= True)
#: E251:7
foo(bar =True)
#: E251:7 E251:9
foo(bar = True)
#: E251:13
y = bar(root= "sdasd")
parser.add_argument('--long-option',
                    #: E135+1:20
                    default=
                    "/rather/long/filesystem/path/here/blah/blah/blah")
parser.add_argument('--long-option',
                    default=
                        "/rather/long/filesystem")
# TODO this looks so stupid.
parser.add_argument('--long-option', default
                    ="/rather/long/filesystem/path/here/blah/blah/blah")
#: E251+2:7 E251+2:9
foo(True,
    baz=(1, 2),
    biz = 'foo'
    )
# Okay
foo(bar=(1 == 1))
foo(bar=(1 != 1))
foo(bar=(1 >= 1))
foo(bar=(1 <= 1))
(options, args) = parser.parse_args()
d[type(None)] = _deepcopy_atomic



import re
import sys
from ast import literal_eval
from functools import total_ordering
from typing import NamedTuple, Sequence, Union

# The following is a list in Python that are line breaks in str.splitlines, but
# not in Python. In Python only \r (Carriage Return, 0xD) and \n (Line Feed,
# 0xA) are allowed to split lines.
_NON_LINE_BREAKS = (
    '\v',  # Vertical Tabulation 0xB
    '\f',  # Form Feed 0xC
    '\x1C',  # File Separator
    '\x1D',  # Group Separator
    '\x1E',  # Record Separator
    '\x85',  # Next Line (NEL - Equivalent to CR+LF.
             # Used to mark end-of-line on some IBM mainframes.)
    '\u2028',  # Line Separator
    '\u2029',  # Paragraph Separator
)


class Version(NamedTuple):
    major: int
    minor: int
    micro: int


def split_lines(string: str, keepends: bool = False) -> Sequence[str]:
    r"""
    Intended for Python code. In contrast to Python's :py:meth:`str.splitlines`,
    looks at form feeds and other special characters as normal text. Just
    splits ``\n`` and ``\r\n``.
    Also different: Returns ``[""]`` for an empty string input.

    In Python 2.7 form feeds are used as normal characters when using
    str.splitlines. However in Python 3 somewhere there was a decision to split
    also on form feeds.
    """
    if keepends:
        lst = string.splitlines(True)

        # We have to merge lines that were broken by form feed characters.
        merge = []
        for i, line in enumerate(lst):
            try:
                last_chr = line[-1]
            except IndexError:
                pass
            else:
                if last_chr in _NON_LINE_BREAKS:
                    merge.append(i)

        for index in reversed(merge):
            try:
                lst[index] = lst[index] + lst[index + 1]
                del lst[index + 1]
            except IndexError:
                # index + 1 can be empty and therefore there's no need to
                # merge.
                pass

        # The stdlib's implementation of the end is inconsistent when calling
        # it with/without keepends. One time there's an empty string in the
        # end, one time there's none.
        if string.endswith('\n') or string.endswith('\r') or string == '':
            lst.append('')
        return lst
    else:
        return re.split(r'\n|\r\n|\r', string)


def python_bytes_to_unicode(
    source: Union[str, bytes], encoding: str = 'utf-8', errors: str = 'strict'
) -> str:
    """
    Checks for unicode BOMs and PEP 263 encoding declarations. Then returns a
    unicode object like in :py:meth:`bytes.decode`.

    :param encoding: See :py:meth:`bytes.decode` documentation.
    :param errors: See :py:meth:`bytes.decode` documentation. ``errors`` can be
        ``'strict'``, ``'replace'`` or ``'ignore'``.
    """
    def detect_encoding():
        """
        For the implementation of encoding definitions in Python, look at:
        - http://www.python.org/dev/peps/pep-0263/
        - http://docs.python.org/2/reference/lexical_analysis.html#encoding-declarations
        """
        byte_mark = literal_eval(r"b'\xef\xbb\xbf'")
        if source.startswith(byte_mark):
            # UTF-8 byte-order mark
            return 'utf-8'

        first_two_lines = re.match(br'(?:[^\r\n]*(?:\r\n|\r|\n)){0,2}', source).group(0)
        possible_encoding = re.search(br"coding[=:]\s*([-\w.]+)",
                                      first_two_lines)
        if possible_encoding:
            e = possible_encoding.group(1)
            if not isinstance(e, str):
                e = str(e, 'ascii', 'replace')
            return e
        else:
            # the default if nothing else has been set -> PEP 263
            return encoding

    if isinstance(source, str):
        # only cast str/bytes
        return source

    encoding = detect_encoding()
    try:
        # Cast to unicode
        return str(source, encoding, errors)
    except LookupError:
        if errors == 'replace':
            # This is a weird case that can happen if the given encoding is not
            # a valid encoding. This usually shouldn't happen with provided
            # encodings, but can happen if somebody uses encoding declarations
            # like `# coding: foo-8`.
            return str(source, 'utf-8', errors)
        raise


def version_info() -> Version:
    """
    Returns a namedtuple of parso's version, similar to Python's
    ``sys.version_info``.
    """
    from parso import __version__
    tupl = re.findall(r'[a-z]+|\d+', __version__)
    return Version(*[x if i == 3 else int(x) for i, x in enumerate(tupl)])


class _PythonVersionInfo(NamedTuple):
    major: int
    minor: int


@total_ordering
class PythonVersionInfo(_PythonVersionInfo):
    def __gt__(self, other):
        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError("Can only compare to tuples of length 2.")
            return (self.major, self.minor) > other
        super().__gt__(other)

        return (self.major, self.minor)

    def __eq__(self, other):
        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError("Can only compare to tuples of length 2.")
            return (self.major, self.minor) == other
        super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)


def _parse_version(version) -> PythonVersionInfo:
    match = re.match(r'(\d+)(?:\.(\d{1,2})(?:\.\d+)?)?((a|b|rc)\d)?$', version)
    if match is None:
        raise ValueError('The given version is not in the right format. '
                         'Use something like "3.8" or "3".')

    major = int(match.group(1))
    minor = match.group(2)
    if minor is None:
        # Use the latest Python in case it's not exactly defined, because the
        # grammars are typically backwards compatible?
        if major == 2:
            minor = "7"
        elif major == 3:
            minor = "6"
        else:
            raise NotImplementedError("Sorry, no support yet for those fancy new/old versions.")
    minor = int(minor)
    return PythonVersionInfo(major, minor)


def parse_version_string(version: str = None) -> PythonVersionInfo:
    """
    Checks for a valid version number (e.g. `3.8` or `3.10.1` or `3`) and
    returns a corresponding version info that is always two characters long in
    decimal.
    """
    if version is None:
        version = '%s.%s' % sys.version_info[:2]
    if not isinstance(version, str):
        raise TypeError('version must be a string like "3.8"')

    return _parse_version(version)



import hashlib
import os
from typing import Generic, TypeVar, Union, Dict, Optional, Any
from pathlib import Path

from parso._compatibility import is_pypy
from parso.pgen2 import generate_grammar
from parso.utils import split_lines, python_bytes_to_unicode, \
    PythonVersionInfo, parse_version_string
from parso.python.diff import DiffParser
from parso.python.tokenize import tokenize_lines, tokenize
from parso.python.token import PythonTokenTypes
from parso.cache import parser_cache, load_module, try_to_save_module
from parso.parser import BaseParser
from parso.python.parser import Parser as PythonParser
from parso.python.errors import ErrorFinderConfig
from parso.python import pep8
from parso.file_io import FileIO, KnownContentFileIO
from parso.normalizer import RefactoringNormalizer, NormalizerConfig

_loaded_grammars: Dict[str, 'Grammar'] = {}

_NodeT = TypeVar("_NodeT")


class Grammar(Generic[_NodeT]):
    """
    :py:func:`parso.load_grammar` returns instances of this class.

    Creating custom none-python grammars by calling this is not supported, yet.

    :param text: A BNF representation of your grammar.
    """
    _start_nonterminal: str
    _error_normalizer_config: Optional[ErrorFinderConfig] = None
    _token_namespace: Any = None
    _default_normalizer_config: NormalizerConfig = pep8.PEP8NormalizerConfig()

    def __init__(self, text: str, *, tokenizer, parser=BaseParser, diff_parser=None):
        self._pgen_grammar = generate_grammar(
            text,
            token_namespace=self._get_token_namespace()
        )
        self._parser = parser
        self._tokenizer = tokenizer
        self._diff_parser = diff_parser
        self._hashed = hashlib.sha256(text.encode("utf-8")).hexdigest()

    def parse(self,
              code: Union[str, bytes] = None,
              *,
              error_recovery=True,
              path: Union[os.PathLike, str] = None,
              start_symbol: str = None,
              cache=False,
              diff_cache=False,
              cache_path: Union[os.PathLike, str] = None,
              file_io: FileIO = None) -> _NodeT:
        """
        If you want to parse a Python file you want to start here, most likely.

        If you need finer grained control over the parsed instance, there will be
        other ways to access it.

        :param str code: A unicode or bytes string. When it's not possible to
            decode bytes to a string, returns a
            :py:class:`UnicodeDecodeError`.
        :param bool error_recovery: If enabled, any code will be returned. If
            it is invalid, it will be returned as an error node. If disabled,
            you will get a ParseError when encountering syntax errors in your
            code.
        :param str start_symbol: The grammar rule (nonterminal) that you want
            to parse. Only allowed to be used when error_recovery is False.
        :param str path: The path to the file you want to open. Only needed for caching.
        :param bool cache: Keeps a copy of the parser tree in RAM and on disk
            if a path is given. Returns the cached trees if the corresponding
            files on disk have not changed. Note that this stores pickle files
            on your file system (e.g. for Linux in ``~/.cache/parso/``).
        :param bool diff_cache: Diffs the cached python module against the new
            code and tries to parse only the parts that have changed. Returns
            the same (changed) module that is found in cache. Using this option
            requires you to not do anything anymore with the cached modules
            under that path, because the contents of it might change. This
            option is still somewhat experimental. If you want stability,
            please don't use it.
        :param bool cache_path: If given saves the parso cache in this
            directory. If not given, defaults to the default cache places on
            each platform.

        :return: A subclass of :py:class:`parso.tree.NodeOrLeaf`. Typically a
            :py:class:`parso.python.tree.Module`.
        """
        if code is None and path is None and file_io is None:
            raise TypeError("Please provide either code or a path.")

        if isinstance(path, str):
            path = Path(path)
        if isinstance(cache_path, str):
            cache_path = Path(cache_path)

        if start_symbol is None:
            start_symbol = self._start_nonterminal

        if error_recovery and start_symbol != 'file_input':
            raise NotImplementedError("This is currently not implemented.")

        if file_io is None:
            if code is None:
                file_io = FileIO(path)  # type: ignore
            else:
                file_io = KnownContentFileIO(path, code)

        if cache and file_io.path is not None:
            module_node = load_module(self._hashed, file_io, cache_path=cache_path)
            if module_node is not None:
                return module_node  # type: ignore

        if code is None:
            code = file_io.read()
        code = python_bytes_to_unicode(code)

        lines = split_lines(code, keepends=True)
        if diff_cache:
            if self._diff_parser is None:
                raise TypeError("You have to define a diff parser to be able "
                                "to use this option.")
            try:
                module_cache_item = parser_cache[self._hashed][file_io.path]
            except KeyError:
                pass
            else:
                module_node = module_cache_item.node
                old_lines = module_cache_item.lines
                if old_lines == lines:
                    return module_node  # type: ignore

                new_node = self._diff_parser(
                    self._pgen_grammar, self._tokenizer, module_node
                ).update(
                    old_lines=old_lines,
                    new_lines=lines
                )
                try_to_save_module(self._hashed, file_io, new_node, lines,
                                   # Never pickle in pypy, it's slow as hell.
                                   pickling=cache and not is_pypy,
                                   cache_path=cache_path)
                return new_node  # type: ignore

        tokens = self._tokenizer(lines)

        p = self._parser(
            self._pgen_grammar,
            error_recovery=error_recovery,
            start_nonterminal=start_symbol
        )
        root_node = p.parse(tokens=tokens)

        if cache or diff_cache:
            try_to_save_module(self._hashed, file_io, root_node, lines,
                               # Never pickle in pypy, it's slow as hell.
                               pickling=cache and not is_pypy,
                               cache_path=cache_path)
        return root_node  # type: ignore

    def _get_token_namespace(self):
        ns = self._token_namespace
        if ns is None:
            raise ValueError("The token namespace should be set.")
        return ns

    def iter_errors(self, node):
        """
        Given a :py:class:`parso.tree.NodeOrLeaf` returns a generator of
        :py:class:`parso.normalizer.Issue` objects. For Python this is
        a list of syntax/indentation errors.
        """
        if self._error_normalizer_config is None:
            raise ValueError("No error normalizer specified for this grammar.")

        return self._get_normalizer_issues(node, self._error_normalizer_config)

    def refactor(self, base_node, node_to_str_map):
        return RefactoringNormalizer(node_to_str_map).walk(base_node)

    def _get_normalizer(self, normalizer_config):
        if normalizer_config is None:
            normalizer_config = self._default_normalizer_config
            if normalizer_config is None:
                raise ValueError("You need to specify a normalizer, because "
                                 "there's no default normalizer for this tree.")
        return normalizer_config.create_normalizer(self)

    def _normalize(self, node, normalizer_config=None):
        """
        TODO this is not public, yet.
        The returned code will be normalized, e.g. PEP8 for Python.
        """
        normalizer = self._get_normalizer(normalizer_config)
        return normalizer.walk(node)

    def _get_normalizer_issues(self, node, normalizer_config=None):
        normalizer = self._get_normalizer(normalizer_config)
        normalizer.walk(node)
        return normalizer.issues

    def __repr__(self):
        nonterminals = self._pgen_grammar.nonterminal_to_dfas.keys()
        txt = ' '.join(list(nonterminals)[:3]) + ' ...'
        return '<%s:%s>' % (self.__class__.__name__, txt)


class PythonGrammar(Grammar):
    _error_normalizer_config = ErrorFinderConfig()
    _token_namespace = PythonTokenTypes
    _start_nonterminal = 'file_input'

    def __init__(self, version_info: PythonVersionInfo, bnf_text: str):
        super().__init__(
            bnf_text,
            tokenizer=self._tokenize_lines,
            parser=PythonParser,
            diff_parser=DiffParser
        )
        self.version_info = version_info

    def _tokenize_lines(self, lines, **kwargs):
        return tokenize_lines(lines, version_info=self.version_info, **kwargs)

    def _tokenize(self, code):
        # Used by Jedi.
        return tokenize(code, version_info=self.version_info)


def load_grammar(*, version: str = None, path: str = None):
    """
    Loads a :py:class:`parso.Grammar`. The default version is the current Python
    version.

    :param str version: A python version string, e.g. ``version='3.8'``.
    :param str path: A path to a grammar file
    """
    version_info = parse_version_string(version)

    file = path or os.path.join(
        'python',
        'grammar%s%s.txt' % (version_info.major, version_info.minor)
    )

    global _loaded_grammars
    path = os.path.join(os.path.dirname(__file__), file)
    try:
        return _loaded_grammars[path]
    except KeyError:
        try:
            with open(path) as f:
                bnf_text = f.read()

            grammar = PythonGrammar(version_info, bnf_text)
            return _loaded_grammars.setdefault(path, grammar)
        except FileNotFoundError:
            message = "Python version %s.%s is currently not supported." % (
                version_info.major, version_info.minor
            )
            raise NotImplementedError(message)



"""
Some syntax errors are a bit complicated and need exact checking. Here we
gather some of the potentially dangerous ones.
"""

from __future__ import division

# With a dot it's not a future import anymore.
from .__future__ import absolute_import

'' ''
''r''u''
b'' BR''


for x in [1]:
    break
    continue

try:
    pass
except ZeroDivisionError:
    pass
    #: E722:0
except:
    pass

try:
    pass
    #: E722:0 E901:0
except:
    pass
except ZeroDivisionError:
    pass


r'\n'
r'\x'
b'\n'


a = 3


def x(b=a):
    global a


def x(*args, c=2, d):
    pass


def x(*, c=2, d):
    pass


def x(a, b=1, *args, c=2, d):
    pass


def x(a, b=1, *, c=2, d):
    pass


lambda *args, c=2, d: (c, d)
lambda *, c=2, d: (c, d)
lambda a, b=1, *args, c=2, d: (c, d)
lambda a, b=1, *, c=2, d: (c, d)


*foo, a = (1,)
*foo[0], a = (1,)
*[], a = (1,)


async def foo():
    await bar()
    #: E901
    yield from []
    return
    #: E901
    return ''


# With decorator it's a different statement.
@bla
async def foo():
    await bar()
    #: E901
    yield from []
    return
    #: E901
    return ''


foo: int = 4
(foo): int = 3
((foo)): int = 3
foo.bar: int
foo[3]: int


def glob():
    global x
    y: foo = x


def c():
    a = 3

    def d():
        class X():
            nonlocal a


def x():
    a = 3

    def y():
        nonlocal a


def x():
    def y():
        nonlocal a

    a = 3


def x():
    a = 3

    def y():
        class z():
            nonlocal a


def x(a):
    def y():
        nonlocal a


def x(a, b):
    def y():
        nonlocal b
        nonlocal a


def x(a):
    def y():
        def z():
            nonlocal a


def x():
    def y(a):
        def z():
            nonlocal a


a = *args, *args
error[(*args, *args)] = 3
*args, *args



# -*- coding: utf-8 -*-
import sys
from textwrap import dedent


def indent(code):
    lines = code.splitlines(True)
    return ''.join([' ' * 2 + line for line in lines])


def build_nested(code, depth, base='def f():\n'):
    if depth == 0:
        return code

    new_code = base + indent(code)
    return build_nested(new_code, depth - 1, base=base)


FAILING_EXAMPLES = [
    '1 +',
    '?',
    'continue',
    'break',
    'return',
    'yield',

    # SyntaxError from Python/ast.c
    'f(x for x in bar, 1)',
    'from foo import a,',
    'from __future__ import whatever',
    'from __future__ import braces',
    'from .__future__ import whatever',
    'def f(x=3, y): pass',
    'lambda x=3, y: x',
    '__debug__ = 1',
    'with x() as __debug__: pass',

    '[]: int',
    '[a, b]: int',
    '(): int',
    '(()): int',
    '((())): int',
    '{}: int',
    'True: int',
    '(a, b): int',
    '*star,: int',
    'a, b: int = 3',
    'foo(+a=3)',
    'f(lambda: 1=1)',
    'f(x=1, x=2)',
    'f(**x, y)',
    'f(x=2, y)',
    'f(**x, *y)',
    'f(**x, y=3, z)',
    # augassign
    'a, b += 3',
    '(a, b) += 3',
    '[a, b] += 3',
    '[a, 1] += 3',
    'f() += 1',
    'lambda x:None+=1',
    '{} += 1',
    '{a:b} += 1',
    '{1} += 1',
    '{*x} += 1',
    '(x,) += 1',
    '(x, y if a else q) += 1',
    '[] += 1',
    '[1,2] += 1',
    '[] += 1',
    'None += 1',
    '... += 1',
    'a > 1 += 1',
    '"test" += 1',
    '1 += 1',
    '1.0 += 1',
    '(yield) += 1',
    '(yield from x) += 1',
    '(x if x else y) += 1',
    'a() += 1',
    'a + b += 1',
    '+a += 1',
    'a and b += 1',
    '*a += 1',
    'a, b += 1',
    'f"xxx" += 1',
    # All assignment tests
    'lambda a: 1 = 1',
    '[x for x in y] = 1',
    '{x for x in y} = 1',
    '{x:x for x in y} = 1',
    '(x for x in y) = 1',
    'None = 1',
    '... = 1',
    'a == b = 1',
    '{a, b} = 1',
    '{a: b} = 1',
    '1 = 1',
    '"" = 1',
    'b"" = 1',
    'b"" = 1',
    '"" "" = 1',
    '1 | 1 = 3',
    '1**1 = 3',
    '~ 1 = 3',
    'not 1 = 3',
    '1 and 1 = 3',
    'def foo(): (yield 1) = 3',
    'def foo(): x = yield 1 = 3',
    'async def foo(): await x = 3',
    '(a if a else a) = a',
    'a, 1 = x',
    'foo() = 1',
    # Cases without the equals but other assignments.
    'with x as foo(): pass',
    'del bar, 1',
    'for x, 1 in []: pass',
    'for (not 1) in []: pass',
    '[x for 1 in y]',
    '[x for a, 3 in y]',
    '(x for 1 in y)',
    '{x for 1 in y}',
    '{x:x for 1 in y}',
    # Unicode/Bytes issues.
    r'u"\x"',
    r'u"\"',
    r'u"\u"',
    r'u"""\U"""',
    r'u"\Uffffffff"',
    r"u'''\N{}'''",
    r"u'\N{foo}'",
    r'b"\x"',
    r'b"\"',
    'b"ä"',

    '*a, *b = 3, 3',
    'async def foo(): yield from []',
    'yield from []',
    '*a = 3',
    'del *a, b',
    'def x(*): pass',
    '(%s *d) = x' % ('a,' * 256),
    '{**{} for a in [1]}',
    '(True,) = x',
    '([False], a) = x',
    'def x(): from math import *',

    # invalid del statements
    'del x + y',
    'del x(y)',
    'async def foo(): del await x',
    'def foo(): del (yield x)',
    'del [x for x in range(10)]',
    'del *x',
    'del *x,',
    'del (*x,)',
    'del [*x]',
    'del x, *y',
    'del *x.y,',
    'del *x[y],',
    'del *x[y::], z',
    'del x, (y, *z)',
    'del (x, *[y, z])',
    'del [x, *(y, [*z])]',
    'del {}',
    'del {x}',
    'del {x, y}',
    'del {x, *y}',

    # invalid starred expressions
    '*x',
    '(*x)',
    '((*x))',
    '1 + (*x)',
    '*x; 1',
    '1; *x',
    '1\n*x',
    'x = *y',
    'x: int = *y',
    'def foo(): return *x',
    'def foo(): yield *x',
    'f"{*x}"',
    'for *x in 1: pass',
    '[1 for *x in 1]',

    # str/bytes combinations
    '"s" b""',
    '"s" b"" ""',
    'b"" "" b"" ""',
    'f"s" b""',
    'b"s" f""',

    # Parser/tokenize.c
    r'"""',
    r'"',
    r"'''",
    r"'",
    r"\blub",
    # IndentationError: too many levels of indentation
    build_nested('pass', 100),

    # SyntaxErrors from Python/symtable.c
    'def f(x, x): pass',
    'nonlocal a',

    # IndentationError
    ' foo',
    'def x():\n    1\n 2',
    'def x():\n 1\n  2',
    'if 1:\nfoo',
    'if 1: blubb\nif 1:\npass\nTrue and False',

    # f-strings
    'f"{}"',
    r'f"{\}"',
    'f"{\'\\\'}"',
    'f"{#}"',
    "f'{1!b}'",
    "f'{1:{5:{3}}}'",
    "f'{'",
    "f'{'",
    "f'}'",
    "f'{\"}'",
    "f'{\"}'",
    # Now nested parsing
    "f'{continue}'",
    "f'{1;1}'",
    "f'{a;}'",
    "f'{b\"\" \"\"}'",
    # f-string expression part cannot include a backslash
    r'''f"{'\n'}"''',

    'async def foo():\n yield x\n return 1',
    'async def foo():\n yield x\n return 1',

    '[*[] for a in [1]]',
    'async def bla():\n def x():  await bla()',
    'del None',
    'del True',
    'del False',
    'del ...',

    # Errors of global / nonlocal
    dedent('''
        def glob():
            x = 3
            x.z
            global x'''),
    dedent('''
        def glob():
            x = 3
            global x'''),
    dedent('''
        def glob():
            x
            global x'''),
    dedent('''
        def glob():
            x = 3
            x.z
            nonlocal x'''),
    dedent('''
        def glob():
            x = 3
            nonlocal x'''),
    dedent('''
        def glob():
            x
            nonlocal x'''),
    # Annotation issues
    dedent('''
        def glob():
            x[0]: foo
            global x'''),
    dedent('''
        def glob():
            x.a: foo
            global x'''),
    dedent('''
        def glob():
            x: foo
            global x'''),
    dedent('''
        def glob():
            x: foo = 5
            global x'''),
    dedent('''
        def glob():
            x: foo = 5
            x
            global x'''),
    dedent('''
        def glob():
            global x
            x: foo = 3
        '''),
    # global/nonlocal + param
    dedent('''
        def glob(x):
            global x
        '''),
    dedent('''
        def glob(x):
            nonlocal x
        '''),
    dedent('''
        def x():
            a =3
            def z():
                nonlocal a
                a = 3
                nonlocal a
        '''),
    dedent('''
        def x():
            a = 4
            def y():
                global a
                nonlocal a
        '''),
    # Missing binding of nonlocal
    dedent('''
        def x():
            nonlocal a
        '''),
    dedent('''
        def x():
            def y():
                nonlocal a
        '''),
    dedent('''
        def x():
            a = 4
            def y():
                global a
                print(a)
                def z():
                    nonlocal a
        '''),
    # Name is assigned before nonlocal declaration
    dedent('''
        def x(a):
            def y():
                a = 10
                nonlocal a
       '''),
]

if sys.version_info[:2] >= (3, 7):
    # This is somehow ok in previous versions.
    FAILING_EXAMPLES += [
        'class X(base for base in bases): pass',
    ]

if sys.version_info[:2] < (3, 8):
    FAILING_EXAMPLES += [
        # Python/compile.c
        dedent('''\
            for a in [1]:
                try:
                    pass
                finally:
                    continue
            '''),  # 'continue' not supported inside 'finally' clause"
    ]

if sys.version_info[:2] >= (3, 8):
    # assignment expressions from issue#89
    FAILING_EXAMPLES += [
        # Case 2
        '(lambda: x := 1)',
        '((lambda: x) := 1)',
        # Case 3
        '(a[i] := x)',
        '((a[i]) := x)',
        '(a(i) := x)',
        # Case 4
        '(a.b := c)',
        '[(i.i:= 0) for ((i), j) in range(5)]',
        # Case 5
        '[i:= 0 for i, j in range(5)]',
        '[(i:= 0) for ((i), j) in range(5)]',
        '[(i:= 0) for ((i), j), in range(5)]',
        '[(i:= 0) for ((i), j.i), in range(5)]',
        '[[(i:= i) for j in range(5)] for i in range(5)]',
        '[i for i, j in range(5) if True or (i:= 1)]',
        '[False and (i:= 0) for i, j in range(5)]',
        # Case 6
        '[i+1 for i in (i:= range(5))]',
        '[i+1 for i in (j:= range(5))]',
        '[i+1 for i in (lambda: (j:= range(5)))()]',
        # Case 7
        'class Example:\n [(j := i) for i in range(5)]',
        # Not in that issue
        '(await a := x)',
        '((await a) := x)',
        # new discoveries
        '((a, b) := (1, 2))',
        '([a, b] := [1, 2])',
        '({a, b} := {1, 2})',
        '({a: b} := {1: 2})',
        '(a + b := 1)',
        '(True := 1)',
        '(False := 1)',
        '(None := 1)',
        '(__debug__ := 1)',
        # Unparenthesized walrus not allowed in dict literals, dict comprehensions and slices
        '{a:="a": b:=1}',
        '{y:=1: 2 for x in range(5)}',
        'a[b:=0:1:2]',
    ]
    # f-string debugging syntax with invalid conversion character
    FAILING_EXAMPLES += [
        "f'{1=!b}'",
    ]



if True:
    result = some_function_that_takes_arguments(
        'a', 'b', 'c',
        'd', 'e', 'f',
        #: E123:0
)
#: E122+1
if some_very_very_very_long_variable_name or var \
or another_very_long_variable_name:
    raise Exception()
#: E122+1
if some_very_very_very_long_variable_name or var[0] \
or another_very_long_variable_name:
    raise Exception()
if True:
    #: E122+1
    if some_very_very_very_long_variable_name or var \
    or another_very_long_variable_name:
        raise Exception()
if True:
    #: E122+1
    if some_very_very_very_long_variable_name or var[0] \
    or another_very_long_variable_name:
        raise Exception()

#: E901+1:8
dictionary = [
    "is": {
        # Might be a E122:4, but is not because the code is invalid Python.
    "nested": yes(),
    },
]
setup('',
      scripts=[''],
      classifiers=[
          #: E121:6
      'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
      ])


#: E123+2:4 E291:15
abc = "E123", (   
    "bad", "hanging", "close"
    )

result = {
    'foo': [
        'bar', {
            'baz': 'frop',
            #: E123
            }
        #: E123
        ]
    #: E123
    }
result = some_function_that_takes_arguments(
    'a', 'b', 'c',
    'd', 'e', 'f',
    #: E123
    )
my_list = [1, 2, 3,
           4, 5, 6,
           #: E124:0
]
my_list = [1, 2, 3,
           4, 5, 6,
           #: E124:19
                   ]
#: E124+2
result = some_function_that_takes_arguments('a', 'b', 'c',
                                            'd', 'e', 'f',
)
fooff(aaaa,
      cca(
          vvv,
          dadd
      ), fff,
      #: E124:0
)
fooff(aaaa,
      ccaaa(
          vvv,
          dadd
      ),
      fff,
      #: E124:0
)
d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s)" % DEFAULT_EXCLUDE
         #: E124:14
              )

if line_removed:
    self.event(cr, uid,
               #: E128:8
        name="Removing the option for contract",
               #: E128:8
        description="contract line has been removed",
               #: E124:8
        )

#: E129+1:4
if foo is None and bar is "frop" and \
    blah == 'yeah':
    blah = 'yeahnah'


#: E129+1:4 E129+2:4
def long_function_name(
    var_one, var_two, var_three,
    var_four):
    hello(var_one)


def qualify_by_address(
        #: E129:4 E129+1:4
    self, cr, uid, ids, context=None,
    params_to_check=frozenset(QUALIF_BY_ADDRESS_PARAM)):
    """ This gets called by the web server """


#: E129+1:4 E129+2:4
if (a == 2 or
    b == "abc def ghi"
    "jkl mno"):
    True

my_list = [
    1, 2, 3,
    4, 5, 6,
    #: E123:8
        ]

abris = 3 + \
        4 + \
        5 + 6

fixed = re.sub(r'\t+', ' ', target[c::-1], 1)[::-1] + \
        target[c + 1:]

rv.update(dict.fromkeys((
              'qualif_nr', 'reasonComment_en', 'reasonComment_fr',
              #: E121:12
            'reasonComment_de', 'reasonComment_it'),
                        '?'),
          #: E128:4
    "foo")
#: E126+1:8
eat_a_dict_a_day({
        "foo": "bar",
})
#: E129+1:4
if (
    x == (
            3
            #: E129:4
    ) or
        y == 4):
    pass
#: E129+1:4 E121+2:8 E129+3:4
if (
    x == (
        3
    ) or
        x == (
            # This one has correct indentation.
            3
            #: E129:4
    ) or
        y == 4):
    pass
troublesome_hash = {
    "hash": "value",
    #: E135+1:8
    "long": "the quick brown fox jumps over the lazy dog before doing a "
        "somersault",
}

# Arguments on first line forbidden when not using vertical alignment
#: E128+1:4
foo = long_function_name(var_one, var_two,
    var_three, var_four)

#: E128+1:4
hello('l.%s\t%s\t%s\t%r' %
    (token[2][0], pos, tokenize.tok_name[token[0]], token[1]))


def qualify_by_address(self, cr, uid, ids, context=None,
                       #: E128:8
        params_to_check=frozenset(QUALIF_BY_ADDRESS_PARAM)):
    """ This gets called by the web server """



import re
import tempfile
import shutil
import logging
import os
from pathlib import Path

import pytest

import parso
from parso import cache
from parso.utils import parse_version_string

collect_ignore = ["setup.py"]

_SUPPORTED_VERSIONS = '3.6', '3.7', '3.8', '3.9', '3.10'


@pytest.fixture(scope='session')
def clean_parso_cache():
    """
    Set the default cache directory to a temporary directory during tests.

    Note that you can't use built-in `tmpdir` and `monkeypatch`
    fixture here because their scope is 'function', which is not used
    in 'session' scope fixture.

    This fixture is activated in ../pytest.ini.
    """
    old = cache._default_cache_path
    tmp = tempfile.mkdtemp(prefix='parso-test-')
    cache._default_cache_path = Path(tmp)
    yield
    cache._default_cache_path = old
    shutil.rmtree(tmp)


def pytest_addoption(parser):
    parser.addoption("--logging", "-L", action='store_true',
                     help="Enables the logging output.")


def pytest_generate_tests(metafunc):
    if 'normalizer_issue_case' in metafunc.fixturenames:
        base_dir = os.path.join(os.path.dirname(__file__), 'test', 'normalizer_issue_files')

        cases = list(colllect_normalizer_tests(base_dir))
        metafunc.parametrize(
            'normalizer_issue_case',
            cases,
            ids=[c.name for c in cases]
        )
    elif 'each_version' in metafunc.fixturenames:
        metafunc.parametrize('each_version', _SUPPORTED_VERSIONS)
    elif 'version_ge_py38' in metafunc.fixturenames:
        ge38 = set(_SUPPORTED_VERSIONS) - {'3.6', '3.7'}
        metafunc.parametrize('version_ge_py38', sorted(ge38))


class NormalizerIssueCase:
    """
    Static Analysis cases lie in the static_analysis folder.
    The tests also start with `#!`, like the goto_definition tests.
    """
    def __init__(self, path):
        self.path = path
        self.name = os.path.basename(path)
        match = re.search(r'python([\d.]+)\.py', self.name)
        self.python_version = match and match.group(1)


def colllect_normalizer_tests(base_dir):
    for f_name in os.listdir(base_dir):
        if f_name.endswith(".py"):
            path = os.path.join(base_dir, f_name)
            yield NormalizerIssueCase(path)


def pytest_configure(config):
    if config.option.logging:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        #ch = logging.StreamHandler(sys.stdout)
        #ch.setLevel(logging.DEBUG)
        #formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        #ch.setFormatter(formatter)

        #root.addHandler(ch)


class Checker:
    def __init__(self, version, is_passing):
        self.version = version
        self._is_passing = is_passing
        self.grammar = parso.load_grammar(version=self.version)

    def parse(self, code):
        if self._is_passing:
            return parso.parse(code, version=self.version, error_recovery=False)
        else:
            self._invalid_syntax(code)

    def _invalid_syntax(self, code):
        with pytest.raises(parso.ParserSyntaxError):
            module = parso.parse(code, version=self.version, error_recovery=False)
            # For debugging
            print(module.children)

    def get_error(self, code):
        errors = list(self.grammar.iter_errors(self.grammar.parse(code)))
        assert bool(errors) != self._is_passing
        if errors:
            return errors[0]

    def get_error_message(self, code):
        error = self.get_error(code)
        if error is None:
            return
        return error.message

    def assert_no_error_in_passing(self, code):
        if self._is_passing:
            module = self.grammar.parse(code)
            assert not list(self.grammar.iter_errors(module))


@pytest.fixture
def works_not_in_py(each_version):
    return Checker(each_version, False)


@pytest.fixture
def works_in_py(each_version):
    return Checker(each_version, True)


@pytest.fixture
def works_ge_py38(each_version):
    version_info = parse_version_string(each_version)
    return Checker(each_version, version_info >= (3, 8))


@pytest.fixture
def works_ge_py39(each_version):
    version_info = parse_version_string(each_version)
    return Checker(each_version, version_info >= (3, 9))




def qualify_by_address(
        self, cr, uid, ids, context=None,
        params_to_check=frozenset(QUALIF_BY_ADDRESS_PARAM)):
    """ This gets called by the web server """


def qualify_by_address(self, cr, uid, ids, context=None,
                       params_to_check=frozenset(QUALIF_BY_ADDRESS_PARAM)):
    """ This gets called by the web server """


_ipv4_re = re.compile('^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.'
                      '(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.'
                      '(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.'
                      '(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$')


fct("""
    AAA """ + status_2_string)


if context:
    msg = """\
action: GET-CONFIG
payload:
    ip_address: "%(ip)s"
    username: "%(username)s"
""" % context


if context:
    msg = """\
action: \
GET-CONFIG
""" % context


if context:
    #: E122+2:0
    msg = """\
action: """\
"""GET-CONFIG
""" % context


def unicode2html(s):
    """Convert the characters &<>'" in string s to HTML-safe sequences.
    Convert newline to <br> too."""
    #: E127+1:28
    return unicode((s or '').replace('&', '&amp;')
                            .replace('\n', '<br>\n'))


parser.add_option('--count', action='store_true',
                  help="print total number of errors and warnings "
                       "to standard error and set exit code to 1 if "
                       "total is not null")

parser.add_option('--exclude', metavar='patterns', default=DEFAULT_EXCLUDE,
                  help="exclude files or directories which match these "
                       "comma separated patterns (default: %s)" %
                       DEFAULT_EXCLUDE)

add_option('--count',
           #: E135+1
           help="print total number of errors "
           "to standard error total is not null")

add_option('--count',
           #: E135+2:11
           help="print total number of errors "
                "to standard error "
           "total is not null")


help = ("print total number of errors " +
        "to standard error")

help = "print total number of errors " \
       "to standard error"

help = u"print total number of errors " \
       u"to standard error"

help = b"print total number of errors " \
       b"to standard error"

#: E122+1:5
help = br"print total number of errors " \
     br"to standard error"

d = dict('foo', help="exclude files or directories which match these "
                     #: E135:9
         "comma separated patterns (default: %s)" % DEFAULT_EXCLUDE)

d = dict('foo', help=u"exclude files or directories which match these "
                     u"comma separated patterns (default: %s)"
                     % DEFAULT_EXCLUDE)

#: E135+1:9 E135+2:9
d = dict('foo', help=b"exclude files or directories which match these "
         b"comma separated patterns (default: %s)"
         % DEFAULT_EXCLUDE)

d = dict('foo', help=br"exclude files or directories which match these "
                     br"comma separated patterns (default: %s)" %
                     DEFAULT_EXCLUDE)

d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s)" %
              DEFAULT_EXCLUDE)

d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s, %s)" %
              (DEFAULT_EXCLUDE, DEFAULT_IGNORE)
         )

d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s, %s)" %
              # who knows what might happen here?
              (DEFAULT_EXCLUDE, DEFAULT_IGNORE)
         )

# parens used to allow the indenting.
troublefree_hash = {
    "hash": "value",
    "long": ("the quick brown fox jumps over the lazy dog before doing a "
             "somersault"),
    "long key that tends to happen more when you're indented": (
        "stringwithalongtoken you don't want to break"
    ),
}

# another accepted form
troublefree_hash = {
    "hash": "value",
    "long": "the quick brown fox jumps over the lazy dog before doing "
            "a somersault",
    ("long key that tends to happen more "
     "when you're indented"): "stringwithalongtoken you don't want to break",
}
# confusing but accepted... don't do that
troublesome_hash = {
    "hash": "value",
    "long": "the quick brown fox jumps over the lazy dog before doing a "
            #: E135:4
    "somersault",
    "longer":
        "the quick brown fox jumps over the lazy dog before doing a "
        "somersaulty",
    "long key that tends to happen more "
    "when you're indented": "stringwithalongtoken you don't want to break",
}

d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s)" %
              DEFAULT_EXCLUDE
         )
d = dict('foo',
         help="exclude files or directories which match these "
              "comma separated patterns (default: %s)" % DEFAULT_EXCLUDE,
         foobar="this clearly should work, because it is at "
                "the right indent level",
         )

rv.update(dict.fromkeys(
              ('qualif_nr', 'reasonComment_en', 'reasonComment_fr',
               'reasonComment_de', 'reasonComment_it'),
              '?'), "foo",
          context={'alpha': 4, 'beta': 53242234, 'gamma': 17})


def f():
    try:
        if not Debug:
            hello('''
If you would like to see debugging output,
try: %s -d5
''' % sys.argv[0])


# The try statement above was not finished.
#: E901
d = {  # comment
    1: 2
}

# issue 138 (we won't allow this in parso)
#: E126+2:9
[
    12,  # this is a multi-line inline
         # comment
]
# issue 151
#: E122+1:3
if a > b and \
   c > d:
    moo_like_a_cow()

my_list = [
    1, 2, 3,
    4, 5, 6,
]

my_list = [1, 2, 3,
           4, 5, 6,
           ]

result = some_function_that_takes_arguments(
    'a', 'b', 'c',
    'd', 'e', 'f',
)

result = some_function_that_takes_arguments('a', 'b', 'c',
                                            'd', 'e', 'f',
                                            )

# issue 203
dica = {
    ('abc'
     'def'): (
        'abc'),
}

(abcdef[0]
 [1]) = (
    'abc')

('abc'
 'def') == (
    'abc')

# issue 214
bar(
    1).zap(
    2)

bar(
    1).zap(
    2)

if True:

    def example_issue254():
        return [node.copy(
                    (
                        replacement
                        # First, look at all the node's current children.
                        for child in node.children
                        # Replace them.
                        for replacement in replace(child)
                    ),
                    dict(name=token.undefined)
                )]


def valid_example():
    return [node.copy(properties=dict(
                          (key, val if val is not None else token.undefined)
                          for key, val in node.items()
                      ))]


foo([
    'bug'
])

# issue 144, finally!
some_hash = {
    "long key that tends to happen more when you're indented":
        "stringwithalongtoken you don't want to break",
}

{
    1:
        999999 if True
        else 0,
}


abc = dedent(
    '''
        mkdir -p ./{build}/
        mv ./build/ ./{build}/%(revision)s/
    '''.format(
        build='build',
        # more stuff
    )
)



"""
Testing if parso finds syntax errors and indentation errors.
"""
import sys
import warnings

import pytest

import parso

from textwrap import dedent
from parso._compatibility import is_pypy
from .failing_examples import FAILING_EXAMPLES, indent, build_nested


if is_pypy:
    # The errors in PyPy might be different. Just skip the module for now.
    pytestmark = pytest.mark.skip()


def _get_error_list(code, version=None):
    grammar = parso.load_grammar(version=version)
    tree = grammar.parse(code)
    return list(grammar.iter_errors(tree))


def assert_comparison(code, error_code, positions):
    errors = [(error.start_pos, error.code) for error in _get_error_list(code)]
    assert [(pos, error_code) for pos in positions] == errors


@pytest.mark.parametrize('code', FAILING_EXAMPLES)
def test_python_exception_matches(code):
    wanted, line_nr = _get_actual_exception(code)

    errors = _get_error_list(code)
    actual = None
    if errors:
        error, = errors
        actual = error.message
    assert actual in wanted
    # Somehow in Python2.7 the SyntaxError().lineno is sometimes None
    assert line_nr is None or line_nr == error.start_pos[0]


def test_non_async_in_async():
    """
    This example doesn't work with FAILING_EXAMPLES, because the line numbers
    are not always the same / incorrect in Python 3.8.
    """
    # Raises multiple errors in previous versions.
    code = 'async def foo():\n def nofoo():[x async for x in []]'
    wanted, line_nr = _get_actual_exception(code)

    errors = _get_error_list(code)
    if errors:
        error, = errors
        actual = error.message
    assert actual in wanted
    if sys.version_info[:2] not in ((3, 8), (3, 9)):
        assert line_nr == error.start_pos[0]
    else:
        assert line_nr == 0  # For whatever reason this is zero in Python 3.8/3.9


@pytest.mark.parametrize(
    ('code', 'positions'), [
        ('1 +', [(1, 3)]),
        ('1 +\n', [(1, 3)]),
        ('1 +\n2 +', [(1, 3), (2, 3)]),
        ('x + 2', []),
        ('[\n', [(2, 0)]),
        ('[\ndef x(): pass', [(2, 0)]),
        ('[\nif 1: pass', [(2, 0)]),
        ('1+?', [(1, 2)]),
        ('?', [(1, 0)]),
        ('??', [(1, 0)]),
        ('? ?', [(1, 0)]),
        ('?\n?', [(1, 0), (2, 0)]),
        ('? * ?', [(1, 0)]),
        ('1 + * * 2', [(1, 4)]),
        ('?\n1\n?', [(1, 0), (3, 0)]),
    ]
)
def test_syntax_errors(code, positions):
    assert_comparison(code, 901, positions)


@pytest.mark.parametrize(
    ('code', 'positions'), [
        (' 1', [(1, 0)]),
        ('def x():\n    1\n 2', [(3, 0)]),
        ('def x():\n 1\n  2', [(3, 0)]),
        ('def x():\n1', [(2, 0)]),
    ]
)
def test_indentation_errors(code, positions):
    assert_comparison(code, 903, positions)


def _get_actual_exception(code):
    with warnings.catch_warnings():
        # We don't care about warnings where locals/globals misbehave here.
        # It's as simple as either an error or not.
        warnings.filterwarnings('ignore', category=SyntaxWarning)
        try:
            compile(code, '<unknown>', 'exec')
        except (SyntaxError, IndentationError) as e:
            wanted = e.__class__.__name__ + ': ' + e.msg
            line_nr = e.lineno
        except ValueError as e:
            # The ValueError comes from byte literals in Python 2 like '\x'
            # that are oddly enough not SyntaxErrors.
            wanted = 'SyntaxError: (value error) ' + str(e)
            line_nr = None
        else:
            assert False, "The piece of code should raise an exception."

    # SyntaxError
    if wanted == 'SyntaxError: assignment to keyword':
        return [wanted, "SyntaxError: can't assign to keyword",
                'SyntaxError: cannot assign to __debug__'], line_nr
    elif wanted == 'SyntaxError: f-string: unterminated string':
        wanted = 'SyntaxError: EOL while scanning string literal'
    elif wanted == 'SyntaxError: f-string expression part cannot include a backslash':
        return [
            wanted,
            "SyntaxError: EOL while scanning string literal",
            "SyntaxError: unexpected character after line continuation character",
        ], line_nr
    elif wanted == "SyntaxError: f-string: expecting '}'":
        wanted = 'SyntaxError: EOL while scanning string literal'
    elif wanted == 'SyntaxError: f-string: empty expression not allowed':
        wanted = 'SyntaxError: invalid syntax'
    elif wanted == "SyntaxError: f-string expression part cannot include '#'":
        wanted = 'SyntaxError: invalid syntax'
    elif wanted == "SyntaxError: f-string: single '}' is not allowed":
        wanted = 'SyntaxError: invalid syntax'
    return [wanted], line_nr


def test_default_except_error_postition():
    # For this error the position seemed to be one line off in Python < 3.10,
    # but that doesn't really matter.
    code = 'try: pass\nexcept: pass\nexcept X: pass'
    wanted, line_nr = _get_actual_exception(code)
    error, = _get_error_list(code)
    assert error.message in wanted
    if sys.version_info[:2] >= (3, 10):
        assert line_nr == error.start_pos[0]
    else:
        assert line_nr != error.start_pos[0]
    # I think this is the better position.
    assert error.start_pos[0] == 2


def test_statically_nested_blocks():
    def build(code, depth):
        if depth == 0:
            return code

        new_code = 'if 1:\n' + indent(code)
        return build(new_code, depth - 1)

    def get_error(depth, add_func=False):
        code = build('foo', depth)
        if add_func:
            code = 'def bar():\n' + indent(code)
        errors = _get_error_list(code)
        if errors:
            assert errors[0].message == 'SyntaxError: too many statically nested blocks'
            return errors[0]
        return None

    assert get_error(19) is None
    assert get_error(19, add_func=True) is None

    assert get_error(20)
    assert get_error(20, add_func=True)


def test_future_import_first():
    def is_issue(code, *args, **kwargs):
        code = code % args
        return bool(_get_error_list(code, **kwargs))

    i1 = 'from __future__ import division'
    i2 = 'from __future__ import absolute_import'
    i3 = 'from __future__ import annotations'
    assert not is_issue(i1)
    assert not is_issue(i1 + ';' + i2)
    assert not is_issue(i1 + '\n' + i2)
    assert not is_issue('"";' + i1)
    assert not is_issue('"";' + i1)
    assert not is_issue('""\n' + i1)
    assert not is_issue('""\n%s\n%s', i1, i2)
    assert not is_issue('""\n%s;%s', i1, i2)
    assert not is_issue('"";%s;%s ', i1, i2)
    assert not is_issue('"";%s\n%s ', i1, i2)
    assert not is_issue(i3, version="3.7")
    assert is_issue(i3, version="3.6")
    assert is_issue('1;' + i1)
    assert is_issue('1\n' + i1)
    assert is_issue('"";1\n' + i1)
    assert is_issue('""\n%s\nfrom x import a\n%s', i1, i2)
    assert is_issue('%s\n""\n%s', i1, i2)


def test_named_argument_issues(works_not_in_py):
    message = works_not_in_py.get_error_message('def foo(*, **dict): pass')
    message = works_not_in_py.get_error_message('def foo(*): pass')
    if works_not_in_py.version.startswith('2'):
        assert message == 'SyntaxError: invalid syntax'
    else:
        assert message == 'SyntaxError: named arguments must follow bare *'

    works_not_in_py.assert_no_error_in_passing('def foo(*, name): pass')
    works_not_in_py.assert_no_error_in_passing('def foo(bar, *, name=1): pass')
    works_not_in_py.assert_no_error_in_passing('def foo(bar, *, name=1, **dct): pass')


def test_escape_decode_literals(each_version):
    """
    We are using internal functions to assure that unicode/bytes escaping is
    without syntax errors. Here we make a bit of quality assurance that this
    works through versions, because the internal function might change over
    time.
    """
    def get_msg(end, to=1):
        base = "SyntaxError: (unicode error) 'unicodeescape' " \
               "codec can't decode bytes in position 0-%s: " % to
        return base + end

    def get_msgs(escape):
        return (get_msg('end of string in escape sequence'),
                get_msg(r"truncated %s escape" % escape))

    error, = _get_error_list(r'u"\x"', version=each_version)
    assert error.message in get_msgs(r'\xXX')

    error, = _get_error_list(r'u"\u"', version=each_version)
    assert error.message in get_msgs(r'\uXXXX')

    error, = _get_error_list(r'u"\U"', version=each_version)
    assert error.message in get_msgs(r'\UXXXXXXXX')

    error, = _get_error_list(r'u"\N{}"', version=each_version)
    assert error.message == get_msg(r'malformed \N character escape', to=2)

    error, = _get_error_list(r'u"\N{foo}"', version=each_version)
    assert error.message == get_msg(r'unknown Unicode character name', to=6)

    # Finally bytes.
    error, = _get_error_list(r'b"\x"', version=each_version)
    wanted = r'SyntaxError: (value error) invalid \x escape at position 0'
    assert error.message == wanted


def test_too_many_levels_of_indentation():
    assert not _get_error_list(build_nested('pass', 99))
    assert _get_error_list(build_nested('pass', 100))
    base = 'def x():\n if x:\n'
    assert not _get_error_list(build_nested('pass', 49, base=base))
    assert _get_error_list(build_nested('pass', 50, base=base))


def test_paren_kwarg():
    assert _get_error_list("print((sep)=seperator)", version="3.8")
    assert not _get_error_list("print((sep)=seperator)", version="3.7")


@pytest.mark.parametrize(
    'code', [
        "f'{*args,}'",
        r'f"\""',
        r'f"\\\""',
        r'fr"\""',
        r'fr"\\\""',
        r"print(f'Some {x:.2f} and some {y}')",
        # Unparenthesized yield expression
        'def foo(): return f"{yield 1}"',
    ]
)
def test_valid_fstrings(code):
    assert not _get_error_list(code, version='3.6')


@pytest.mark.parametrize(
    'code', [
        'a = (b := 1)',
        '[x4 := x ** 5 for x in range(7)]',
        '[total := total + v for v in range(10)]',
        'while chunk := file.read(2):\n pass',
        'numbers = [y := math.factorial(x), y**2, y**3]',
        '{(a:="a"): (b:=1)}',
        '{(y:=1): 2 for x in range(5)}',
        'a[(b:=0)]',
        'a[(b:=0, c:=0)]',
        'a[(b:=0):1:2]',
    ]
)
def test_valid_namedexpr(code):
    assert not _get_error_list(code, version='3.8')


@pytest.mark.parametrize(
    'code', [
        '{x := 1, 2, 3}',
        '{x4 := x ** 5 for x in range(7)}',
    ]
)
def test_valid_namedexpr_set(code):
    assert not _get_error_list(code, version='3.9')


@pytest.mark.parametrize(
    'code', [
        'a[b:=0]',
        'a[b:=0, c:=0]',
    ]
)
def test_valid_namedexpr_index(code):
    assert not _get_error_list(code, version='3.10')


@pytest.mark.parametrize(
    ('code', 'message'), [
        ("f'{1+}'", ('invalid syntax')),
        (r'f"\"', ('invalid syntax')),
        (r'fr"\"', ('invalid syntax')),
    ]
)
def test_invalid_fstrings(code, message):
    """
    Some fstring errors are handled differntly in 3.6 and other versions.
    Therefore check specifically for these errors here.
    """
    error, = _get_error_list(code, version='3.6')
    assert message in error.message


@pytest.mark.parametrize(
    'code', [
        "from foo import (\nbar,\n rab,\n)",
        "from foo import (bar, rab, )",
    ]
)
def test_trailing_comma(code):
    errors = _get_error_list(code)
    assert not errors


def test_continue_in_finally():
    code = dedent('''\
        for a in [1]:
            try:
                pass
            finally:
                continue
        ''')
    assert not _get_error_list(code, version="3.8")
    assert _get_error_list(code, version="3.7")


@pytest.mark.parametrize(
    'template', [
        "a, b, {target}, c = d",
        "a, b, *{target}, c = d",
        "(a, *{target}), c = d",
        "for x, {target} in y: pass",
        "for x, q, {target} in y: pass",
        "for x, q, *{target} in y: pass",
        "for (x, *{target}), q in y: pass",
    ]
)
@pytest.mark.parametrize(
    'target', [
        "True",
        "False",
        "None",
        "__debug__"
    ]
)
def test_forbidden_name(template, target):
    assert _get_error_list(template.format(target=target), version="3")


def test_repeated_kwarg():
    # python 3.9+ shows which argument is repeated
    assert (
        _get_error_list("f(q=1, q=2)", version="3.8")[0].message
        == "SyntaxError: keyword argument repeated"
    )
    assert (
        _get_error_list("f(q=1, q=2)", version="3.9")[0].message
        == "SyntaxError: keyword argument repeated: q"
    )


@pytest.mark.parametrize(
    ('source', 'no_errors'), [
        ('a(a for a in b,)', False),
        ('a(a for a in b, a)', False),
        ('a(a, a for a in b)', False),
        ('a(a, b, a for a in b, c, d)', False),
        ('a(a for a in b)', True),
        ('a((a for a in b), c)', True),
        ('a(c, (a for a in b))', True),
        ('a(a, b, (a for a in b), c, d)', True),
    ]
)
def test_unparenthesized_genexp(source, no_errors):
    assert bool(_get_error_list(source)) ^ no_errors


@pytest.mark.parametrize(
    ('source', 'no_errors'), [
        ('*x = 2', False),
        ('(*y) = 1', False),
        ('((*z)) = 1', False),
        ('*a,', True),
        ('*a, = 1', True),
        ('(*a,)', True),
        ('(*a,) = 1', True),
        ('[*a]', True),
        ('[*a] = 1', True),
        ('a, *b', True),
        ('a, *b = 1', True),
        ('a, *b, c', True),
        ('a, *b, c = 1', True),
        ('a, (*b, c), d', True),
        ('a, (*b, c), d = 1', True),
        ('*a.b,', True),
        ('*a.b, = 1', True),
        ('*a[b],', True),
        ('*a[b], = 1', True),
        ('*a[b::], c', True),
        ('*a[b::], c = 1', True),
        ('(a, *[b, c])', True),
        ('(a, *[b, c]) = 1', True),
        ('[a, *(b, [*c])]', True),
        ('[a, *(b, [*c])] = 1', True),
        ('[*(1,2,3)]', True),
        ('{*(1,2,3)}', True),
        ('[*(1,2,3),]', True),
        ('[*(1,2,3), *(4,5,6)]', True),
        ('[0, *(1,2,3)]', True),
        ('{*(1,2,3),}', True),
        ('{*(1,2,3), *(4,5,6)}', True),
        ('{0, *(4,5,6)}', True)
    ]
)
def test_starred_expr(source, no_errors):
    assert bool(_get_error_list(source, version="3")) ^ no_errors


@pytest.mark.parametrize(
    'code', [
        'a, (*b), c',
        'a, (*b), c = 1',
        'a, ((*b)), c',
        'a, ((*b)), c = 1',
    ]
)
def test_parenthesized_single_starred_expr(code):
    assert not _get_error_list(code, version='3.8')
    assert _get_error_list(code, version='3.9')


@pytest.mark.parametrize(
    'code', [
        '() = ()',
        '() = []',
        '[] = ()',
        '[] = []',
    ]
)
def test_valid_empty_assignment(code):
    assert not _get_error_list(code)


@pytest.mark.parametrize(
    'code', [
        'del ()',
        'del []',
        'del x',
        'del x,',
        'del x, y',
        'del (x, y)',
        'del [x, y]',
        'del (x, [y, z])',
        'del x.y, x[y]',
        'del f(x)[y::]',
        'del x[[*y]]',
        'del x[[*y]::]',
    ]
)
def test_valid_del(code):
    assert not _get_error_list(code)


@pytest.mark.parametrize(
    ('source', 'version', 'no_errors'), [
        ('[x for x in range(10) if lambda: 1]', '3.8', True),
        ('[x for x in range(10) if lambda: 1]', '3.9', False),
        ('[x for x in range(10) if (lambda: 1)]', '3.9', True),
    ]
)
def test_lambda_in_comp_if(source, version, no_errors):
    assert bool(_get_error_list(source, version=version)) ^ no_errors



"""
These tests test the cases that the old fast parser tested with the normal
parser.

The old fast parser doesn't exist anymore and was replaced with a diff parser.
However the tests might still be relevant for the parser.
"""

from textwrap import dedent

from parso import parse


def test_carriage_return_splitting():
    source = dedent('''



        "string"

        class Foo():
            pass
        ''')
    source = source.replace('\n', '\r\n')
    module = parse(source)
    assert [n.value for lst in module.get_used_names().values() for n in lst] == ['Foo']


def check_p(src, number_parsers_used, number_of_splits=None, number_of_misses=0):
    if number_of_splits is None:
        number_of_splits = number_parsers_used

    module_node = parse(src)

    assert src == module_node.get_code()
    return module_node


def test_for():
    src = dedent("""\
    for a in [1,2]:
        a

    for a1 in 1,"":
        a1
    """)
    check_p(src, 1)


def test_class_with_class_var():
    src = dedent("""\
    class SuperClass:
        class_super = 3
        def __init__(self):
            self.foo = 4
    pass
    """)
    check_p(src, 3)


def test_func_with_if():
    src = dedent("""\
    def recursion(a):
        if foo:
            return recursion(a)
        else:
            if bar:
                return inexistent
            else:
                return a
    """)
    check_p(src, 1)


def test_decorator():
    src = dedent("""\
    class Decorator():
        @memoize
        def dec(self, a):
            return a
    """)
    check_p(src, 2)


def test_nested_funcs():
    src = dedent("""\
    def memoize(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    """)
    check_p(src, 3)


def test_multi_line_params():
    src = dedent("""\
    def x(a,
          b):
        pass

    foo = 1
    """)
    check_p(src, 2)


def test_class_func_if():
    src = dedent("""\
    class Class:
        def func(self):
            if 1:
                a
            else:
                b

    pass
    """)
    check_p(src, 3)


def test_multi_line_for():
    src = dedent("""\
    for x in [1,
              2]:
        pass

    pass
    """)
    check_p(src, 1)


def test_wrong_indentation():
    src = dedent("""\
    def func():
        a
         b
        a
    """)
    check_p(src, 1)

    src = dedent("""\
    def complex():
        def nested():
            a
             b
            a

        def other():
            pass
    """)
    check_p(src, 3)


def test_strange_parentheses():
    src = dedent("""
    class X():
        a = (1
    if 1 else 2)
        def x():
            pass
    """)
    check_p(src, 2)


def test_fake_parentheses():
    """
    The fast parser splitting counts parentheses, but not as correct tokens.
    Therefore parentheses in string tokens are included as well. This needs to
    be accounted for.
    """
    src = dedent(r"""
    def x():
        a = (')'
    if 1 else 2)
        def y():
            pass
        def z():
            pass
    """)
    check_p(src, 3, 2, 1)


def test_additional_indent():
    source = dedent('''\
    int(
      def x():
          pass
    ''')

    check_p(source, 2)


def test_round_trip():
    code = dedent('''
    def x():
        """hahaha"""
    func''')

    assert parse(code).get_code() == code


def test_parentheses_in_string():
    code = dedent('''
    def x():
        '('

    import abc

    abc.''')
    check_p(code, 2, 1, 1)



for a in 'abc':
    for b in 'xyz':
        hello(a)  # indented with 8 spaces
        #: E903:0
	hello(b)  # indented with 1 tab
if True:
    #: E101:0
	pass

#: E122+1
change_2_log = \
"""Change 2 by slamb@testclient on 2006/04/13 21:46:23

	creation
"""

p4change = {
    2: change_2_log,
}


class TestP4Poller(unittest.TestCase):
    def setUp(self):
        self.setUpGetProcessOutput()
        return self.setUpChangeSource()

    def tearDown(self):
        pass


#
if True:
    #: E101:0 E101+1:0
	foo(1,
	    2)


def test_keys(self):
    """areas.json - All regions are accounted for."""
    expected = set([
        #: E101:0
	u'Norrbotten',
        #: E101:0
	u'V\xe4sterbotten',
    ])


if True:
    hello("""
	tab at start of this line
""")



#!/usr/bin/env python

from __future__ import with_statement

from setuptools import setup, find_packages

import parso


__AUTHOR__ = 'David Halter'
__AUTHOR_EMAIL__ = 'davidhalter88@gmail.com'

readme = open('README.rst').read() + '\n\n' + open('CHANGELOG.rst').read()

setup(
    name='parso',
    version=parso.__version__,
    description='A Python Parser',
    author=__AUTHOR__,
    author_email=__AUTHOR_EMAIL__,
    include_package_data=True,
    maintainer=__AUTHOR__,
    maintainer_email=__AUTHOR_EMAIL__,
    url='https://github.com/davidhalter/parso',
    license='MIT',
    keywords='python parser parsing',
    long_description=readme,
    packages=find_packages(exclude=['test']),
    package_data={'parso': ['python/grammar*.txt', 'py.typed', '*.pyi', '**/*.pyi']},
    platforms=['any'],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Editors :: Integrated Development Environments (IDE)',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    extras_require={
        'testing': [
            'pytest<6.0.0',
            'docopt',
        ],
        'qa': [
            'flake8==3.8.3',
            'mypy==0.782',
        ],
    },
)



abc = "E121", (
    #: E121:2
  "dent")
abc = "E122", (
    #: E121:0
"dent")
my_list = [
    1, 2, 3,
    4, 5, 6,
    #: E123
    ]
abc = "E124", ("visual",
               "indent_two"
               #: E124:14
              )
abc = "E124", ("visual",
               "indent_five"
               #: E124:0
)
a = (123,
     #: E124:0
)
#: E129+1:4
if (row < 0 or self.moduleCount <= row or
    col < 0 or self.moduleCount <= col):
    raise Exception("%s,%s - %s" % (row, col, self.moduleCount))

abc = "E126", (
    #: E126:12
            "dent")
abc = "E126", (
    #: E126:8
        "dent")
abc = "E127", ("over-",
               #: E127:18
                  "over-indent")
abc = "E128", ("visual",
               #: E128:4
    "hanging")
abc = "E128", ("under-",
               #: E128:14
              "under-indent")


my_list = [
    1, 2, 3,
    4, 5, 6,
    #: E123:5
     ]
result = {
    #: E121:3
   'key1': 'value',
    #: E121:3
   'key2': 'value',
}
rv.update(dict.fromkeys((
              'qualif_nr', 'reasonComment_en', 'reasonComment_fr',
              'reasonComment_de', 'reasonComment_it'),
                        #: E128:10
          '?'),
          "foo")

abricot = 3 + \
          4 + \
          5 + 6
abc = "hello", (

    "there",
    #: E126:5
     # "john",
    "dude")
part = set_mimetype((
    a.get('mime_type', 'text')),
                    'default')
part = set_mimetype((
    a.get('mime_type', 'text')),
                    #: E127:21
                     'default')



from textwrap import dedent

import pytest

from parso import parse
# Using star import for easier eval testing below.
from parso.python.tree import *  # noqa: F403
from parso.tree import *  # noqa: F403
from parso.tree import ErrorLeaf, TypedLeaf


@pytest.mark.parametrize(
    'indent,expected_dump', [
        (None, "Module(["
               "Lambda(["
               "Keyword('lambda', (1, 0)), "
               "Param(["
               "Name('x', (1, 7), prefix=' '), "
               "Operator(',', (1, 8)), "
               "]), "
               "Param(["
               "Name('y', (1, 10), prefix=' '), "
               "]), "
               "Operator(':', (1, 11)), "
               "PythonNode('arith_expr', ["
               "Name('x', (1, 13), prefix=' '), "
               "Operator('+', (1, 15), prefix=' '), "
               "Name('y', (1, 17), prefix=' '), "
               "]), "
               "]), "
               "EndMarker('', (1, 18)), "
               "])"),
        (0, dedent('''\
            Module([
            Lambda([
            Keyword('lambda', (1, 0)),
            Param([
            Name('x', (1, 7), prefix=' '),
            Operator(',', (1, 8)),
            ]),
            Param([
            Name('y', (1, 10), prefix=' '),
            ]),
            Operator(':', (1, 11)),
            PythonNode('arith_expr', [
            Name('x', (1, 13), prefix=' '),
            Operator('+', (1, 15), prefix=' '),
            Name('y', (1, 17), prefix=' '),
            ]),
            ]),
            EndMarker('', (1, 18)),
            ])''')),
        (4, dedent('''\
            Module([
                Lambda([
                    Keyword('lambda', (1, 0)),
                    Param([
                        Name('x', (1, 7), prefix=' '),
                        Operator(',', (1, 8)),
                    ]),
                    Param([
                        Name('y', (1, 10), prefix=' '),
                    ]),
                    Operator(':', (1, 11)),
                    PythonNode('arith_expr', [
                        Name('x', (1, 13), prefix=' '),
                        Operator('+', (1, 15), prefix=' '),
                        Name('y', (1, 17), prefix=' '),
                    ]),
                ]),
                EndMarker('', (1, 18)),
            ])''')),
        ('\t', dedent('''\
            Module([
            \tLambda([
            \t\tKeyword('lambda', (1, 0)),
            \t\tParam([
            \t\t\tName('x', (1, 7), prefix=' '),
            \t\t\tOperator(',', (1, 8)),
            \t\t]),
            \t\tParam([
            \t\t\tName('y', (1, 10), prefix=' '),
            \t\t]),
            \t\tOperator(':', (1, 11)),
            \t\tPythonNode('arith_expr', [
            \t\t\tName('x', (1, 13), prefix=' '),
            \t\t\tOperator('+', (1, 15), prefix=' '),
            \t\t\tName('y', (1, 17), prefix=' '),
            \t\t]),
            \t]),
            \tEndMarker('', (1, 18)),
            ])''')),
    ]
)
def test_dump_parser_tree(indent, expected_dump):
    code = "lambda x, y: x + y"
    module = parse(code)
    assert module.dump(indent=indent) == expected_dump

    # Check that dumped tree can be eval'd to recover the parser tree and original code.
    recovered_code = eval(expected_dump).get_code()
    assert recovered_code == code


@pytest.mark.parametrize(
    'node,expected_dump,expected_code', [
        (  # Dump intermediate node (not top level module)
            parse("def foo(x, y): return x + y").children[0], dedent('''\
                Function([
                    Keyword('def', (1, 0)),
                    Name('foo', (1, 4), prefix=' '),
                    PythonNode('parameters', [
                        Operator('(', (1, 7)),
                        Param([
                            Name('x', (1, 8)),
                            Operator(',', (1, 9)),
                        ]),
                        Param([
                            Name('y', (1, 11), prefix=' '),
                        ]),
                        Operator(')', (1, 12)),
                    ]),
                    Operator(':', (1, 13)),
                    ReturnStmt([
                        Keyword('return', (1, 15), prefix=' '),
                        PythonNode('arith_expr', [
                            Name('x', (1, 22), prefix=' '),
                            Operator('+', (1, 24), prefix=' '),
                            Name('y', (1, 26), prefix=' '),
                        ]),
                    ]),
                ])'''),
            "def foo(x, y): return x + y",
        ),
        (  # Dump leaf
            parse("def foo(x, y): return x + y").children[0].children[0],
            "Keyword('def', (1, 0))",
            'def',
        ),
        (  # Dump ErrorLeaf
            ErrorLeaf('error_type', 'error_code', (1, 1), prefix=' '),
            "ErrorLeaf('error_type', 'error_code', (1, 1), prefix=' ')",
            ' error_code',
        ),
        (  # Dump TypedLeaf
            TypedLeaf('type', 'value', (1, 1)),
            "TypedLeaf('type', 'value', (1, 1))",
            'value',
        ),
    ]
)
def test_dump_parser_tree_not_top_level_module(node, expected_dump, expected_code):
    dump_result = node.dump()
    assert dump_result == expected_dump

    # Check that dumped tree can be eval'd to recover the parser tree and original code.
    recovered_code = eval(dump_result).get_code()
    assert recovered_code == expected_code


def test_dump_parser_tree_invalid_args():
    module = parse("lambda x, y: x + y")

    with pytest.raises(TypeError):
        module.dump(indent=1.1)


def test_eval_dump_recovers_parent():
    module = parse("lambda x, y: x + y")
    module2 = eval(module.dump())
    assert module2.parent is None
    lambda_node = module2.children[0]
    assert lambda_node.parent is module2
    assert module2.children[1].parent is module2
    assert lambda_node.children[0].parent is lambda_node
    param_node = lambda_node.children[1]
    assert param_node.parent is lambda_node
    assert param_node.children[0].parent is param_node
    assert param_node.children[1].parent is param_node
    arith_expr_node = lambda_node.children[-1]
    assert arith_expr_node.parent is lambda_node
    assert arith_expr_node.children[0].parent is arith_expr_node



from textwrap import dedent

import pytest

from parso import load_grammar
from parso import ParserSyntaxError
from parso.pgen2 import generate_grammar
from parso.python import tokenize


def _parse(code, version=None):
    code = dedent(code) + "\n\n"
    grammar = load_grammar(version=version)
    return grammar.parse(code, error_recovery=False)


def _invalid_syntax(code, version=None, **kwargs):
    with pytest.raises(ParserSyntaxError):
        module = _parse(code, version=version, **kwargs)
        # For debugging
        print(module.children)


def test_formfeed(each_version):
    s = "foo\n\x0c\nfoo\n"
    t = _parse(s, each_version)
    assert t.children[0].children[0].type == 'name'
    assert t.children[1].children[0].type == 'name'
    s = "1\n\x0c\x0c\n2\n"
    t = _parse(s, each_version)

    with pytest.raises(ParserSyntaxError):
        s = "\n\x0c2\n"
        _parse(s, each_version)


def test_matrix_multiplication_operator(works_in_py):
    works_in_py.parse("a @ b")
    works_in_py.parse("a @= b")


def test_yield_from(works_in_py, each_version):
    works_in_py.parse("yield from x")
    works_in_py.parse("(yield from x) + y")
    _invalid_syntax("yield from", each_version)


def test_await_expr(works_in_py):
    works_in_py.parse("""async def foo():
                         await x
                  """)

    works_in_py.parse("""async def foo():

        def foo(): pass

        def foo(): pass

        await x
    """)

    works_in_py.parse("""async def foo(): return await a""")

    works_in_py.parse("""def foo():
        def foo(): pass
        async def foo(): await x
    """)


@pytest.mark.parametrize(
    'code', [
        "async = 1",
        "await = 1",
        "def async(): pass",
    ]
)
def test_async_var(works_not_in_py, code):
    works_not_in_py.parse(code)


def test_async_for(works_in_py):
    works_in_py.parse("async def foo():\n async for a in b: pass")


@pytest.mark.parametrize("body", [
    """[1 async for a in b
    ]""",
    """[1 async
    for a in b
    ]""",
    """[
    1
    async for a in b
    ]""",
    """[
    1
    async for a
    in b
    ]""",
    """[
    1
    async
    for
    a
    in
    b
    ]""",
    """  [
    1 async for a in b
  ]""",
])
def test_async_for_comprehension_newline(works_in_py, body):
    # Issue #139
    works_in_py.parse("""async def foo():
    {}""".format(body))


def test_async_with(works_in_py):
    works_in_py.parse("async def foo():\n async with a: pass")


def test_async_with_invalid(works_in_py):
    works_in_py.parse("""def foo():\n async with a: pass""")


def test_raise_3x_style_1(each_version):
    _parse("raise", each_version)


def test_raise_2x_style_2(works_not_in_py):
    works_not_in_py.parse("raise E, V")


def test_raise_2x_style_3(works_not_in_py):
    works_not_in_py.parse("raise E, V, T")


def test_raise_2x_style_invalid_1(each_version):
    _invalid_syntax("raise E, V, T, Z", version=each_version)


def test_raise_3x_style(works_in_py):
    works_in_py.parse("raise E1 from E2")


def test_raise_3x_style_invalid_1(each_version):
    _invalid_syntax("raise E, V from E1", each_version)


def test_raise_3x_style_invalid_2(each_version):
    _invalid_syntax("raise E from E1, E2", each_version)


def test_raise_3x_style_invalid_3(each_version):
    _invalid_syntax("raise from E1, E2", each_version)


def test_raise_3x_style_invalid_4(each_version):
    _invalid_syntax("raise E from", each_version)


# Adapted from Python 3's Lib/test/test_grammar.py:GrammarTests.testFuncdef
def test_annotation_1(works_in_py):
    works_in_py.parse("""def f(x) -> list: pass""")


def test_annotation_2(works_in_py):
    works_in_py.parse("""def f(x:int): pass""")


def test_annotation_3(works_in_py):
    works_in_py.parse("""def f(*x:str): pass""")


def test_annotation_4(works_in_py):
    works_in_py.parse("""def f(**x:float): pass""")


def test_annotation_5(works_in_py):
    works_in_py.parse("""def f(x, y:1+2): pass""")


def test_annotation_6(each_version):
    _invalid_syntax("""def f(a, (b:1, c:2, d)): pass""", each_version)


def test_annotation_7(each_version):
    _invalid_syntax("""def f(a, (b:1, c:2, d), e:3=4, f=5, *g:6): pass""", each_version)


def test_annotation_8(each_version):
    s = """def f(a, (b:1, c:2, d), e:3=4, f=5,
                    *g:6, h:7, i=8, j:9=10, **k:11) -> 12: pass"""
    _invalid_syntax(s, each_version)


def test_except_new(each_version):
    s = dedent("""
        try:
            x
        except E as N:
            y""")
    _parse(s, each_version)


def test_except_old(works_not_in_py):
    s = dedent("""
        try:
            x
        except E, N:
            y""")
    works_not_in_py.parse(s)


# Adapted from Python 3's Lib/test/test_grammar.py:GrammarTests.testAtoms
def test_set_literal_1(works_in_py):
    works_in_py.parse("""x = {'one'}""")


def test_set_literal_2(works_in_py):
    works_in_py.parse("""x = {'one', 1,}""")


def test_set_literal_3(works_in_py):
    works_in_py.parse("""x = {'one', 'two', 'three'}""")


def test_set_literal_4(works_in_py):
    works_in_py.parse("""x = {2, 3, 4,}""")


def test_new_octal_notation(each_version):
    _parse("""0o7777777777777""", each_version)
    _invalid_syntax("""0o7324528887""", each_version)


def test_old_octal_notation(works_not_in_py):
    works_not_in_py.parse("07")


def test_long_notation(works_not_in_py):
    works_not_in_py.parse("0xFl")
    works_not_in_py.parse("0xFL")
    works_not_in_py.parse("0b1l")
    works_not_in_py.parse("0B1L")
    works_not_in_py.parse("0o7l")
    works_not_in_py.parse("0O7L")
    works_not_in_py.parse("0l")
    works_not_in_py.parse("0L")
    works_not_in_py.parse("10l")
    works_not_in_py.parse("10L")


def test_new_binary_notation(each_version):
    _parse("""0b101010""", each_version)
    _invalid_syntax("""0b0101021""", each_version)


def test_class_new_syntax(works_in_py):
    works_in_py.parse("class B(t=7): pass")
    works_in_py.parse("class B(t, *args): pass")
    works_in_py.parse("class B(t, **kwargs): pass")
    works_in_py.parse("class B(t, *args, **kwargs): pass")
    works_in_py.parse("class B(t, y=9, *args, **kwargs): pass")


def test_parser_idempotency_extended_unpacking(works_in_py):
    """A cut-down version of pytree_idempotency.py."""
    works_in_py.parse("a, *b, c = x\n")
    works_in_py.parse("[*a, b] = x\n")
    works_in_py.parse("(z, *y, w) = m\n")
    works_in_py.parse("for *z, m in d: pass\n")


def test_multiline_bytes_literals(each_version):
    s = """
        md5test(b"\xaa" * 80,
                (b"Test Using Larger Than Block-Size Key "
                 b"and Larger Than One Block-Size Data"),
                "6f630fad67cda0ee1fb1f562db3aa53e")
        """
    _parse(s, each_version)


def test_multiline_bytes_tripquote_literals(each_version):
    s = '''
        b"""
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN">
        """
        '''
    _parse(s, each_version)


def test_ellipsis(works_in_py, each_version):
    works_in_py.parse("...")
    _parse("[0][...]", version=each_version)


def test_dict_unpacking(works_in_py):
    works_in_py.parse("{**dict(a=3), foo:2}")


def test_multiline_str_literals(each_version):
    s = """
        md5test("\xaa" * 80,
                ("Test Using Larger Than Block-Size Key "
                 "and Larger Than One Block-Size Data"),
                "6f630fad67cda0ee1fb1f562db3aa53e")
        """
    _parse(s, each_version)


def test_py2_backticks(works_not_in_py):
    works_not_in_py.parse("`1`")


def test_py2_string_prefixes(works_not_in_py):
    works_not_in_py.parse("ur'1'")
    works_not_in_py.parse("Ur'1'")
    works_not_in_py.parse("UR'1'")
    _invalid_syntax("ru'1'", works_not_in_py.version)


def py_br(each_version):
    _parse('br""', each_version)


def test_py3_rb(works_in_py):
    works_in_py.parse("rb'1'")
    works_in_py.parse("RB'1'")


def test_left_recursion():
    with pytest.raises(ValueError, match='left recursion'):
        generate_grammar('foo: foo NAME\n', tokenize.PythonTokenTypes)


@pytest.mark.parametrize(
    'grammar, error_match', [
        ['foo: bar | baz\nbar: NAME\nbaz: NAME\n',
         r"foo is ambiguous.*given a (PythonTokenTypes\.)?NAME.*bar or baz"],
        ['''foo: bar | baz\nbar: 'x'\nbaz: "x"\n''',
         r"foo is ambiguous.*given a ReservedString\(x\).*bar or baz"],
        ['''foo: bar | 'x'\nbar: 'x'\n''',
         r"foo is ambiguous.*given a ReservedString\(x\).*bar or foo"],
        # An ambiguity with the second (not the first) child of a production
        ['outer: "a" [inner] "b" "c"\ninner: "b" "c" [inner]\n',
         r"outer is ambiguous.*given a ReservedString\(b\).*inner or outer"],
        # An ambiguity hidden by a level of indirection (middle)
        ['outer: "a" [middle] "b" "c"\nmiddle: inner\ninner: "b" "c" [inner]\n',
         r"outer is ambiguous.*given a ReservedString\(b\).*middle or outer"],
    ]
)
def test_ambiguities(grammar, error_match):
    with pytest.raises(ValueError, match=error_match):
        generate_grammar(grammar, tokenize.PythonTokenTypes)



# -*- coding: utf-8    # This file contains Unicode characters.

from textwrap import dedent

import pytest

from parso import parse
from parso.python import tree
from parso.tree import search_ancestor


class TestsFunctionAndLambdaParsing:

    FIXTURES = [
        ('def my_function(x, y, z) -> str:\n    return x + y * z\n', {
            'name': 'my_function',
            'call_sig': 'my_function(x, y, z)',
            'params': ['x', 'y', 'z'],
            'annotation': "str",
        }),
        ('lambda x, y, z: x + y * z\n', {
            'name': '<lambda>',
            'call_sig': '<lambda>(x, y, z)',
            'params': ['x', 'y', 'z'],
        }),
    ]

    @pytest.fixture(params=FIXTURES)
    def node(self, request):
        parsed = parse(dedent(request.param[0]), version='3.10')
        request.keywords['expected'] = request.param[1]
        child = parsed.children[0]
        if child.type == 'simple_stmt':
            child = child.children[0]
        return child

    @pytest.fixture()
    def expected(self, request, node):
        return request.keywords['expected']

    def test_name(self, node, expected):
        if node.type != 'lambdef':
            assert isinstance(node.name, tree.Name)
            assert node.name.value == expected['name']

    def test_params(self, node, expected):
        assert isinstance(node.get_params(), list)
        assert all(isinstance(x, tree.Param) for x in node.get_params())
        assert [str(x.name.value) for x in node.get_params()] == [x for x in expected['params']]

    def test_is_generator(self, node, expected):
        assert node.is_generator() is expected.get('is_generator', False)

    def test_yields(self, node, expected):
        assert node.is_generator() == expected.get('yields', False)

    def test_annotation(self, node, expected):
        expected_annotation = expected.get('annotation', None)
        if expected_annotation is None:
            assert node.annotation is None
        else:
            assert node.annotation.value == expected_annotation


def test_end_pos_line(each_version):
    # jedi issue #150
    s = "x()\nx( )\nx(  )\nx (  )\n"

    module = parse(s, version=each_version)
    for i, simple_stmt in enumerate(module.children[:-1]):
        expr_stmt = simple_stmt.children[0]
        assert expr_stmt.end_pos == (i + 1, i + 3)


def test_default_param(each_version):
    func = parse('def x(foo=42): pass', version=each_version).children[0]
    param, = func.get_params()
    assert param.default.value == '42'
    assert param.annotation is None
    assert not param.star_count


def test_annotation_param(each_version):
    func = parse('def x(foo: 3): pass', version=each_version).children[0]
    param, = func.get_params()
    assert param.default is None
    assert param.annotation.value == '3'
    assert not param.star_count


def test_annotation_params(each_version):
    func = parse('def x(foo: 3, bar: 4): pass', version=each_version).children[0]
    param1, param2 = func.get_params()

    assert param1.default is None
    assert param1.annotation.value == '3'
    assert not param1.star_count

    assert param2.default is None
    assert param2.annotation.value == '4'
    assert not param2.star_count


def test_default_and_annotation_param(each_version):
    func = parse('def x(foo:3=42): pass', version=each_version).children[0]
    param, = func.get_params()
    assert param.default.value == '42'
    assert param.annotation.value == '3'
    assert not param.star_count


def get_yield_exprs(code, version):
    return list(parse(code, version=version).children[0].iter_yield_exprs())


def get_return_stmts(code):
    return list(parse(code).children[0].iter_return_stmts())


def get_raise_stmts(code, child):
    return list(parse(code).children[child].iter_raise_stmts())


def test_yields(each_version):
    y, = get_yield_exprs('def x(): yield', each_version)
    assert y.value == 'yield'
    assert y.type == 'keyword'

    y, = get_yield_exprs('def x(): (yield 1)', each_version)
    assert y.type == 'yield_expr'

    y, = get_yield_exprs('def x(): [1, (yield)]', each_version)
    assert y.type == 'keyword'


def test_yield_from():
    y, = get_yield_exprs('def x(): (yield from 1)', '3.8')
    assert y.type == 'yield_expr'


def test_returns():
    r, = get_return_stmts('def x(): return')
    assert r.value == 'return'
    assert r.type == 'keyword'

    r, = get_return_stmts('def x(): return 1')
    assert r.type == 'return_stmt'


def test_raises():
    code = """
def single_function():
    raise Exception
def top_function():
    def inner_function():
        raise NotImplementedError()
    inner_function()
    raise Exception
def top_function_three():
    try:
        raise NotImplementedError()
    except NotImplementedError:
        pass
    raise Exception
    """

    r = get_raise_stmts(code, 0)  # Lists in a simple Function
    assert len(list(r)) == 1

    r = get_raise_stmts(code, 1)  # Doesn't Exceptions list in closures
    assert len(list(r)) == 1

    r = get_raise_stmts(code, 2)  # Lists inside try-catch
    assert len(list(r)) == 2


@pytest.mark.parametrize(
    'code, name_index, is_definition, include_setitem', [
        ('x = 3', 0, True, False),
        ('x.y = 3', 0, False, False),
        ('x.y = 3', 1, True, False),
        ('x.y = u.v = z', 0, False, False),
        ('x.y = u.v = z', 1, True, False),
        ('x.y = u.v = z', 2, False, False),
        ('x.y = u.v, w = z', 3, True, False),
        ('x.y = u.v, w = z', 4, True, False),
        ('x.y = u.v, w = z', 5, False, False),

        ('x, y = z', 0, True, False),
        ('x, y = z', 1, True, False),
        ('x, y = z', 2, False, False),
        ('x, y = z', 2, False, False),
        ('x[0], y = z', 2, False, False),
        ('x[0] = z', 0, False, False),
        ('x[0], y = z', 0, False, False),
        ('x[0], y = z', 2, False, True),
        ('x[0] = z', 0, True, True),
        ('x[0], y = z', 0, True, True),
        ('x: int = z', 0, True, False),
        ('x: int = z', 1, False, False),
        ('x: int = z', 2, False, False),
        ('x: int', 0, True, False),
        ('x: int', 1, False, False),
    ]
)
def test_is_definition(code, name_index, is_definition, include_setitem):
    module = parse(code, version='3.8')
    name = module.get_first_leaf()
    while True:
        if name.type == 'name':
            if name_index == 0:
                break
            name_index -= 1
        name = name.get_next_leaf()

    assert name.is_definition(include_setitem=include_setitem) == is_definition


def test_iter_funcdefs():
    code = dedent('''
        def normal(): ...
        async def asyn(): ...
        @dec
        def dec_normal(): ...
        @dec1
        @dec2
        async def dec_async(): ...
        def broken
        ''')
    module = parse(code, version='3.8')
    func_names = [f.name.value for f in module.iter_funcdefs()]
    assert func_names == ['normal', 'asyn', 'dec_normal', 'dec_async']


def test_with_stmt_get_test_node_from_name():
    code = "with A as X.Y, B as (Z), C as Q[0], D as Q['foo']: pass"
    with_stmt = parse(code, version='3').children[0]
    tests = [
        with_stmt.get_test_node_from_name(name).value
        for name in with_stmt.get_defined_names(include_setitem=True)
    ]
    assert tests == ["A", "B", "C", "D"]


sample_module = parse('x + y')
sample_node = sample_module.children[0]
sample_leaf = sample_node.children[0]


@pytest.mark.parametrize(
    'node,node_types,expected_ancestor', [
        (sample_module, ('file_input',), None),
        (sample_node, ('arith_expr',), None),
        (sample_node, ('file_input', 'eval_input'), sample_module),
        (sample_leaf, ('name',), None),
        (sample_leaf, ('arith_expr',), sample_node),
        (sample_leaf, ('file_input',), sample_module),
        (sample_leaf, ('file_input', 'arith_expr'), sample_node),
        (sample_leaf, ('shift_expr',), None),
        (sample_leaf, ('name', 'shift_expr',), None),
        (sample_leaf, (), None),
    ]
)
def test_search_ancestor(node, node_types, expected_ancestor):
    assert node.search_ancestor(*node_types) is expected_ancestor
    assert search_ancestor(node, *node_types) is expected_ancestor  # deprecated
